-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local im  = ui_imgui

local ffi = require('ffi')

local C = {}

C.name = 'Pacenotes'
C.description = 'Plays a string of pacenotes'
C.pinSchema = {
  {dir = 'in', type = 'flow', name = 'flow', description = 'Inflow for this node.'},
  {dir = 'in', type = 'number', name = 'vehID', description = 'Vehicle id reference for computing direction'},
  {dir = 'in', type = 'number', name = 'exitAngle', description = 'The angle at which the vehicle should be travelling when hitting the trigger'}
}

C.tags = {}

local function rotatePoint(point, pivot, angle)
    local radAngle = angle * 0.01745329252
    
    local s = math.sin(radAngle)
    local c = math.cos(radAngle)

    local px = point.x - pivot.x
    local py = point.y - pivot.y

    local pxNew = (px * c - py * s) + pivot.x
    local pyNew = (px * s + py * c) + pivot.y
    
    return { x = pxNew, y = pyNew}  
end

function C:getPosition()
    if self.pinIn.flow == nil then return end
    
    -- get connected flow node
    local node = self.pinIn.flow.node
    if node == nil then return nil end

    -- check node type
    if node.nodeType ~= "events/customizedTriggerBox" then return nil end

    -- get transform pinIn on node
    local positionIn = node.pinIn.position
    if positionIn.empty then return nil end
    
    local transformNode = positionIn.node
    if transformNode.nodeType ~= "types/transform" then return nil end
    
    return transformNode.position
end

function C:getFwdDir()
    local pos = self:getPosition()
    if pos == nil then return nil end

    local globalFwd = {x = 0, y = 1}
    local globalOrigin = {x = 0, y = 0}

    local localFwdAngle = self.pinIn.exitAngle.value or 0
    local localFwd = rotatePoint(globalFwd, globalOrigin, localFwdAngle)

    return vec3(localFwd.x, localFwd.y, 0)
end

function C:init()
  self.data.notes = ""
  self.data.directional = true
end

function C:drawCustomProperties()
  if im.Button("Test") then
    self:work()
  end

  local pos = self:getPosition()
  if pos ~= nil then
      if im.Button("Teleport To") then
          local camRot = quat(getCameraQuat())
          setCameraPosRot(
           pos.x, pos.y, pos.z,
           camRot.x, camRot.y, camRot.z, camRot.w)
      end
  end
end

function C:drawMiddle(builder, style)
  --draw imgui
  builder:Middle()
  im.Text(tostring(self.data.notes))
  if self.data.directional then im.Text("Directional") else im.Text("Any Direction") end

  --draw debug
  local pos = self:getPosition()
  if pos ~= nil then
    local camPos = getCameraPosition()
    local camDist = pos:squaredDistance(camPos)
    if camDist < 25000 then
      debugDrawer:drawLine((pos):toPoint3F(), (pos + vec3(0, 0, 10)):toPoint3F(),ColorF(0,0,1,0.5))
  
      if self.data.directional then
        local fwdDir = self:getFwdDir()
        debugDrawer:drawLine((pos):toPoint3F(), ((fwdDir * 5) + pos):toPoint3F(),ColorF(1,1,0,0.5))    
      end
  
      if self.data.notes ~= "" then
          debugDrawer:drawText((pos + vec3(0, 0, 11)):toPoint3F(), String(self.data.notes), ColorF(1,0,0,1))
      end
    end
  end
end

function C:work()
  local pacenotesExtension = extensions.scripts_pacenotes_extension
  if pacenotesExtension == nil or self.data.notes == "" then return end

  -- check if car is going the wrong way
  local wrongWay = false

  if self.data.directional then
    local fwdDir = self:getFwdDir()
    if self.pinIn.vehID.value and fwdDir ~= nil then
      local veh = scenetree.findObjectById(self.pinIn.vehID.value)
      if veh ~= nil then
        local vehicleData = map.objects[veh:getID()]
        if vehicleData then
           local vehicleDir =  vehicleData.vel
           local dotProduct = vehicleDir:dot(fwdDir)
           wrongWay = dotProduct < 0.25
        end
      end
    end
  end

  -- handle wrong way appropriately
  if wrongWay then
    pacenotesExtension.addPacenotes("wrong way")
  else
    pacenotesExtension.addPacenotes(self.data.notes)
  end
end

return _flowgraph_createNode(C)
