-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

-- info
local logTag = "pacenotes"
local debugLevel = 0

local function setDebugLevel(level)
    debugLevel = level
end

-- sound settings
local defaultSoundData = {
                            volume = 7,
                            pitch = 1,
                            fadeInTime = -1,
                            fadeOutTime = -1,
                            unique = false
                        }
local soundClipAmount = 0.1
local soundPath = "art/sound/pacenotes/"

-- main table
local pacenoteTable = nil

-- QUEUE
-- Queue format : 
-- "sound" key specifies a key name from pacenoteTable, 
-- "time" is the remaining play time, once it hits 0 this is removed from the queue
-- "played" specifies if the oneshot has been played yet
local soundQueue = {}

-- UI Queue format:
-- "icon" key specifies icon (or nil)
-- "type" key specifies background type (or nil)
local uiQueue = {}

-- Debug
local function countTable(tbl)
    local count = 0
    for k,v in pairs(tbl) do
        count = count + 1
    end
    return count
end

local function info()
    local verbose = verbose or false

    local soundFileTable = {}
    log('I', logTag, tostring(countTable(pacenoteTable)) .. ' unique pace notes')
    for k,v in pairs(pacenoteTable) do
        if soundFileTable[v.file] == nil then soundFileTable[v.file] = v.file end
    end
    log('I', logTag, tostring(countTable(soundFileTable)) .. ' different audio files')
end

-- Supplementary helpers

-- This will find "left" or "right" in preveious words
-- for supplementary notes such as "tightens into 2"
-- and automatically queue a "left 2" in its place
local function findTurnSupplementNote(noteIndex, wordTable)
    --uh oh
    if noteIndex <= 1 or wordTable == nil then
        return nil
    end

   local originNote = wordTable[noteIndex]
    for i=noteIndex-1, 1, -1 do
      local word = wordTable[i]
      local lfr =  string.find(word, "left")
      if lfr ~= nil then
        return originNote .. " left" 
      end
      local rfr = string.find(word, "right")
      if rfr ~= nil then
        return originNote .. " right"
      end
    end
    return nil
end

-- Ui queue
local function updateUiQueue()
    if #uiQueue == 0 then return end

    local queuedUi = uiQueue[1]
    if queuedUi.icon ~= nil then
        local iconType = queuedUi.type or "normal"
        local iconImage = queuedUi.icon
        guihooks.trigger('Pacenote', {type=iconType, image=iconImage})
    end
    table.remove(uiQueue, 1)
end


-- Audio queue
local function updateQueue(dt)
    -- nuffin
    if #soundQueue == 0 then return end

    -- get first item 
    local queuedSound = soundQueue[1]

    -- play if not played already, otherwise update normally
    if not queuedSound.played then
            local soundData = pacenoteTable[queuedSound.sound]
            local res = Engine.Audio.playOnce('AudioGui', soundPath .. soundData.file, defaultSoundData)
            if res ~= nil then
                queuedSound.time = math.max(0, res.len - soundClipAmount) + (soundData.extraLength or 0)
                --print("played " .. soundData.file .. ". len = " .. tostring(queuedSound.time))
            else
                queuedSound.time = 0
                log('E', logTag, 'Cannot find sound file ' .. soundData.file .. '. Skipping!!')
            end
            queuedSound.played = true
    else 
        queuedSound.time = queuedSound.time - dt
        if queuedSound.time <= 0 then
            table.remove(soundQueue, 1)
        end
    end
end

local function queueNoteUi(key, noteIndex, wordTable)
    if debugLevel >= 2 then
        log('I', logTag, 'Pacenotes: Queuing ui ' .. key .. ',' .. tostring(noteIndex) .. ',' .. tostring(wordTable))
    end

    local data = pacenoteTable[key]
    if data == nil then
        log('E', logTag, 'queueNoteUi MISSING KEY ' .. tostring(key))
        return
    end

    if data.uiData ~= nil then
        if data.uiData.supplementType ~= nil then
            if data.uiData.supplementType == "turn" then
                local newNote = findTurnSupplementNote(noteIndex, wordTable)
                if newNote ~= nil then
                    queueNoteUi(newNote)
                end
            else
                log('E', logTag, 'Invalid supplement type: ' .. data.uiData.supplementType)
            end
        else
            local uiQueueInfo = data.uiData
            table.insert(uiQueue, uiQueueInfo)
        end
    end
end

local function queueNote(key)
    if debugLevel == 1 or debugLevel > 2 then
        log('I', logTag, 'Pacenotes: Queuing audio ' .. key)
    end

    local soundQueueInfo = {}
    soundQueueInfo.played = false
    soundQueueInfo.sound = key
    table.insert(soundQueue, soundQueueInfo)
end

local function queueNotes(notes)
    if notes == nil then
        log('E', logTag, 'queueNotes got nil notes??')
        return
    end

    -- first split this into words
    local words = {}
    for w in notes:gmatch("%S+") do table.insert(words, w) end

    local i = 1
    while i <= #words do
        local canPeek = #words - i
        if canPeek > 3 then canPeek = 3 end --only peek ahead by 3

        -- try to peek ahead by 3 and try these keys
        local tryKey = nil
        if canPeek > 0 then
            -- move up each word and grab the next few words each time
            for j=canPeek,0,-1 do
                tryKey = ""
                for k=j, 0, -1 do
                    tryKey = words[k+i] .. tryKey
                    if k ~= 0 then tryKey = " " .. tryKey end
                end

                -- check if this is in the table, if so, break out of this loop
                if pacenoteTable[tryKey] ~= nil then
                     i = i + j -- skip our peeked values
                     break 
                end
            end
        else
            tryKey = words[i]
        end

        -- get the associated sound data and queue
        local soundData = pacenoteTable[tryKey]
        if soundData == nil then
            log('E', logTag, 'Cannot find sound key ' .. tryKey .. '. Skipping!!')
        else
            queueNote(tryKey)
            queueNoteUi(tryKey, i, words)
        end

        -- increment counter
        i = i + 1
    end
end

local function onUpdate(dtReal, dtSim)
  updateQueue(dtReal)
  updateUiQueue()
end

-- initialization
local function processAliases()
    -- create a second table to not mess up the following iterator
    local aliasTable = {}

    for k,v in pairs(pacenoteTable) do
        if v.aliases ~= nil then
            for k2, v2 in pairs(v.aliases) do
                local alias = v2
                if aliasTable[alias] ~= nil then
                    log('E', logTag, 'Cannot process alias ' .. v2 .. '. Key already exists.')
                else
                    --set alias to v(pacenote)
                    aliasTable[v2] = v
                end
            end
        end
    end

    -- apply aliases
    for k,v in pairs(aliasTable) do
        pacenoteTable[k] = v
    end
end

local function reload()
  pacenoteTable = readJsonFile("scripts/pacenotes/pacenotes.json") or {}
  processAliases()
end

local function onExtensionLoaded()
    reload()
end

-- public interface
M.onExtensionLoaded = onExtensionLoaded
M.onUpdate = onUpdate

M.setDebugLevel = setDebugLevel
M.addPacenotes = queueNotes
M.info = info
M.reload = reload

return M